"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const zlib = require("zlib");
const cors = require("sfn-cors");
const values = require("lodash/values");
const index_1 = require("../bootstrap/index");
const HttpController_1 = require("../controllers/HttpController");
const StatusException_1 = require("../tools/StatusException");
const functions_1 = require("../tools/functions");
const internal_1 = require("../tools/internal");
const error_1 = require("../tools/internal/error");
const interfaces_1 = require("../tools/interfaces");
const symbols_1 = require("../tools/symbols");
const RouteMap_1 = require("../tools/RouteMap");
const literal_toolkit_1 = require("literal-toolkit");
const check_iterable_1 = require("check-iterable");
const EFFECT_METHODS = [
    "DELETE",
    "PATCH",
    "POST",
    "PUT"
];
const XMLType = /(text|application)\/xml\b/;
index_1.router.onerror = function onerror(err, req, res) {
    res.keepAlive = false;
    let ctrl = new HttpController_1.HttpController(req, res);
    if (res.statusCode === 404)
        err = new StatusException_1.StatusException(404);
    else if (res.statusCode === 405)
        err = new StatusException_1.StatusException(405);
    else if (err instanceof Error)
        err = new StatusException_1.StatusException(500, err.message);
    else if (typeof err === "string")
        err = new StatusException_1.StatusException(500, err);
    else
        err = new StatusException_1.StatusException(500);
    handleError(err, ctrl, req.method + " " + req.shortUrl);
};
function getRouteHandler(key) {
    return async (req, res) => {
        let mod = RouteMap_1.routeMap.resolve(key), methods = RouteMap_1.routeMap.methods(key), ctrl = null, initiated = false;
        res.on("error", (err) => {
            error_1.tryLogError(err, req.method + " " + req.shortUrl);
        });
        try {
            ctrl = mod.create(req, res);
            for (let method of methods) {
                if (!internal_1.isOwnMethod(ctrl, method)) {
                    RouteMap_1.routeMap.del(key, method);
                    continue;
                }
                else if (!initiated) {
                    if (!cors(ctrl.cors, req, res)) {
                        throw new StatusException_1.StatusException(410);
                    }
                    else if (req.method === "OPTIONS") {
                        res.end();
                    }
                    handleCsrfToken(ctrl);
                    await applyInit(ctrl);
                    initiated = true;
                    res.gzip = req.encodings.includes("gzip") && ctrl.gzip;
                    if (req.method == "GET" && ctrl.jsonp && req.query[ctrl.jsonp]) {
                        res.jsonp = req.query[ctrl.jsonp];
                    }
                }
                let result = await ctrl[method](...await getArguments(ctrl, method));
                if (check_iterable_1.isIterableIterator(result) || check_iterable_1.isAsyncIterableIterator(result)) {
                    await handleIteratorResponse(ctrl, result);
                }
                else {
                    await handleResponse(ctrl, result);
                }
            }
            if (initiated) {
                if (!req.isEventSource && !res.finished) {
                    res.end();
                }
                await applyDestroy(ctrl);
                finish(ctrl);
            }
        }
        catch (err) {
            ctrl = ctrl || new HttpController_1.HttpController(req, res);
            handleError(err, ctrl);
        }
    };
}
exports.getRouteHandler = getRouteHandler;
async function applyInit(ctrl) {
    if (typeof ctrl.init === "function") {
        await ctrl.init();
    }
    else if (typeof ctrl.before === "function") {
        await ctrl.before();
    }
}
exports.applyInit = applyInit;
async function applyDestroy(ctrl) {
    if (typeof ctrl.destroy === "function") {
        await ctrl.destroy();
    }
    else if (typeof ctrl.after === "function") {
        await ctrl.after();
    }
}
exports.applyDestroy = applyDestroy;
function handleFinish(err, ctrl, stack) {
    error_1.tryLogError(err, stack);
    finish(ctrl);
}
async function handleError(err, ctrl, stack) {
    let { req, res } = ctrl;
    if (res.sent)
        return handleFinish(err, ctrl, stack);
    let _err;
    if (err instanceof StatusException_1.StatusException && err.code === 405 && req.isEventSource)
        _err = new StatusException_1.StatusException(204);
    else
        _err = StatusException_1.StatusException.from(err);
    if (req.accept == "application/json" || res.jsonp) {
        res.send(ctrl.error(_err.message, _err.code));
    }
    else {
        res.status = _err.code;
        try {
            let { httpErrorView } = ctrl.ctor;
            let content = await httpErrorView(_err, ctrl);
            res.type = "text/html";
            res.send(content);
        }
        catch (err) {
            res.type = "text/plain";
            res.send(err.message);
        }
    }
    handleFinish(err, ctrl, stack);
}
async function getArguments(ctrl, method) {
    let { req, res } = ctrl, data = values(req.params), args = [];
    let meta = Reflect.getMetadata("design:paramtypes", ctrl, method);
    for (let type of meta) {
        if (type === Number) {
            args.push(literal_toolkit_1.number.parse(data.shift()));
        }
        else if (type === Boolean) {
            let val = data.shift();
            args.push(val == "false" || val == "0" ? false : true);
        }
        else if (type === interfaces_1.Request) {
            args.push(req);
        }
        else if (type === interfaces_1.Response) {
            args.push(res);
        }
        else if (type === interfaces_1.Session) {
            args.push(req.session);
        }
        else {
            args.push(data.shift());
        }
    }
    return args;
}
async function handleResponse(ctrl, data) {
    let { req, res } = ctrl;
    if (req.isEventSource) {
        if (data !== undefined)
            ctrl.sse.send(data);
    }
    else if (!res.sent) {
        if (req.method === "HEAD") {
            res.end();
        }
        else if (data !== undefined) {
            let type = res.type;
            if (data === null) {
                res.end("");
            }
            else if (typeof data === "object" && type && XMLType.test(type)) {
                res.xml(data);
            }
            else if (typeof data === "string" && res.gzip) {
                await handleGzip(ctrl, data);
            }
            else {
                res.send(data);
            }
        }
    }
}
async function handleIteratorResponse(ctrl, iter) {
    let { req, res, sse } = ctrl;
    let reqType = req.type;
    let value, done;
    while ({ value, done } = await iter.next()) {
        if (value === undefined) {
            if (done)
                break;
            else
                continue;
        }
        else if (!res.headersSent && !req.isEventSource) {
            if (!res.type && reqType && reqType !== "*/*") {
                res.type = reqType;
            }
            else if (Buffer.isBuffer(value)) {
                res.type = "application/octet-stream";
            }
            else if (typeof value === "string") {
                res.type = "text/plain";
            }
            res.writeHead(200);
        }
        if (req.isEventSource) {
            sse.send(value);
        }
        else if (Buffer.isBuffer(value)) {
            res.write(value);
        }
        else {
            res.write(String(value));
        }
    }
}
async function handleGzip(ctrl, data) {
    let { res } = ctrl;
    data = await new Promise((resolve, reject) => {
        zlib.gzip(data, (err, data) => err ? reject(err) : resolve(data));
    });
    res.headers["content-encoding"] = "gzip";
    res.type = "text/html";
    res.end(data);
}
function finish(ctrl) {
    ctrl.emit("finish", ctrl.req, ctrl.res);
}
function handleCsrfToken(ctrl) {
    if (!ctrl.csrfProtection || !ctrl.req.session)
        return;
    let { req, res } = ctrl;
    if (req.method == "GET") {
        if (!req.session.csrfTokens)
            req.session.csrfTokens = {};
        Object.defineProperty(req, "csrfToken", {
            set(v) {
                req[symbols_1.realCsrfToken] = v;
                req.session.csrfTokens[req.url] = v;
                res.set("X-CSRF-Token", v);
            },
            get() {
                return !req[symbols_1.realCsrfToken] || (req.csrfToken = functions_1.randStr(64));
            }
        });
    }
    else if (EFFECT_METHODS.includes(req.method)) {
        if (!req.headers.referer) {
            throw new StatusException_1.StatusException(403);
        }
        let ref = req.headers.referer, tokens = req.session.csrfTokens, token = tokens && tokens[ref];
        if (token === undefined || (req.headers["x-csrf-token"] != token
            && req.query["x-csrf-token"] != token
            && req.body["x-csrf-token"] != token)) {
            throw new StatusException_1.StatusException(403);
        }
        else {
            Object.defineProperty(req, "csrfToken", {
                set(v) {
                    if (v === null || v === undefined)
                        delete tokens[ref];
                },
                get() {
                    return tokens[ref];
                }
            });
        }
    }
}
//# sourceMappingURL=http-route.js.map