"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const index_1 = require("../bootstrap/index");
const StatusException_1 = require("../tools/StatusException");
const interfaces_1 = require("../tools/interfaces");
const symbols_1 = require("../tools/symbols");
const WebSocketController_1 = require("../controllers/WebSocketController");
const http_init_1 = require("./http-init");
const websocket_init_1 = require("./websocket-init");
const websocket_cookie_1 = require("./websocket-cookie");
const websocket_session_1 = require("./websocket-session");
const internal_1 = require("../tools/internal");
const error_1 = require("../tools/internal/error");
const RouteMap_1 = require("../tools/RouteMap");
const thenable_generator_1 = require("thenable-generator");
const last = require("lodash/last");
const http_route_1 = require("./http-route");
let importedNamesapces = [];
function tryImport(nsp) {
    if (importedNamesapces.includes(nsp))
        return;
    importedNamesapces.push(nsp);
    index_1.ws.of(nsp)
        .use(websocket_init_1.default)
        .use(websocket_cookie_1.default)
        .use(websocket_cookie_1.handler2)
        .use(websocket_session_1.default)
        .use(websocket_session_1.handler2)
        .on("connection", (socket) => {
        for (let item of RouteMap_1.eventMap.values()) {
            socket.on(item.route, (...data) => {
                handleEvent(RouteMap_1.eventMap.keyFor(item), socket, data);
            });
        }
        socket.on("error", (err) => {
            let ctrl = new WebSocketController_1.WebSocketController(socket);
            handleError(err, {
                time: Date.now(),
                event: "",
                code: 500
            }, ctrl, socket.protocol.toUpperCase() + " id: " + socket.id);
        });
    });
}
exports.tryImport = tryImport;
async function handleEvent(key, socket, data) {
    let mod = RouteMap_1.eventMap.resolve(key), methods = RouteMap_1.eventMap.methods(key), { prefix: nsp, route: event } = RouteMap_1.eventMap.get(key), ctrl = null, initiated = false, info = { time: Date.now(), event, code: 200 };
    try {
        socket[symbols_1.activeEvent] = nsp + (last(nsp) == "/" ? "" : "/") + event;
        ctrl = mod.create(socket);
        for (let method of methods) {
            if (!internal_1.isOwnMethod(ctrl, method)) {
                RouteMap_1.eventMap.del(key, method);
                continue;
            }
            else if (!initiated) {
                if (socket.disconnected)
                    return;
                await http_route_1.applyInit(ctrl);
                initiated = true;
            }
            let generator = new thenable_generator_1.ThenableAsyncGenerator(ctrl[method](...getArguments(ctrl, method, data)));
            let value, done;
            while ({ value, done } = await generator.next()) {
                (value === undefined) || socket.emit(event, value);
                if (done) {
                    break;
                }
            }
        }
        if (initiated) {
            await http_route_1.applyDestroy(ctrl);
            finish(ctrl, info);
        }
    }
    catch (err) {
        ctrl = ctrl || new WebSocketController_1.WebSocketController(socket);
        await handleError(err, info, ctrl);
    }
}
function getArguments(ctrl, method, data) {
    let args = [];
    let meta = Reflect.getMetadata("design:paramtypes", ctrl, method);
    for (let type of meta) {
        if (type === interfaces_1.WebSocket) {
            args.push(ctrl.socket);
        }
        else if (type === interfaces_1.Session) {
            args.push(ctrl.socket.session);
        }
        else {
            args.push(data.shift());
        }
    }
    return args;
}
function finish(ctrl, info) {
    let socket = ctrl.socket;
    ctrl.emit("finish", socket);
    http_init_1.logRequest(info.time, socket.protocol.toUpperCase(), info.code, info.event);
}
async function handleError(err, info, ctrl, stack) {
    let _err = StatusException_1.StatusException.from(err);
    info.code = _err.code;
    if (info.event) {
        ctrl.socket.emit(info.event, ctrl.error(_err.message, _err.code));
    }
    error_1.tryLogError(err, stack);
    finish(ctrl, info);
}
//# sourceMappingURL=websocket-event.js.map