"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ipchannel_1 = require("ipchannel");
const values = require("lodash/values");
class MessageChannel {
    constructor(name) {
        this.name = name;
        this.events = {};
        this.clusterListenerMap = new Map();
    }
    publish(event, data, servers) {
        event = this.name + "#" + event;
        if (app.rpc.server) {
            return app.rpc.server.publish(event, data, servers);
        }
        else if (!servers) {
            return ipchannel_1.default.to("all").emit(event, data);
        }
        else {
            let ok = false;
            for (let receiver of servers) {
                receiver = receiver.match(/\d+/)[0];
                ok = true;
                ipchannel_1.default.to(parseInt(receiver)).emit(event, data);
            }
            return ok;
        }
    }
    subscribe(event, listener) {
        event = this.name + "#" + event;
        if (!this.events[event]) {
            this.events[event] = [];
        }
        this.events[event].push(listener);
        for (let connection of values(app.rpc.connections)) {
            connection.subscribe(event, listener);
        }
        if (!app.rpc.server) {
            ipchannel_1.default.on(event, this.bindClusterChannel(listener));
        }
        return this;
    }
    unsubscribe(event, listener) {
        event = this.name + "#" + event;
        let clusterListener;
        let listeners = this.events[event] || [];
        for (let connection of values(app.rpc.connections)) {
            connection.unsubscribe(event, listener);
        }
        if (listener) {
            ipchannel_1.default.removeAllListeners(event);
        }
        else if (clusterListener = this.clusterListenerMap.get(listener)) {
            ipchannel_1.default.removeListener(event, clusterListener);
        }
        if (!listeners.length) {
            return false;
        }
        else if (listener) {
            return listeners.splice(listeners.indexOf(listener), 1).length > 0;
        }
        else {
            return delete this.events[event];
        }
    }
    linkRpcChannel(connection) {
        for (let event in this.events) {
            for (let listener of this.events[event]) {
                connection.subscribe(event, listener);
            }
        }
    }
    bindClusterChannel(listener) {
        let fn = (_, data) => listener(data);
        this.clusterListenerMap.set(listener, fn);
        return fn;
    }
}
exports.MessageChannel = MessageChannel;
class Message {
    constructor(data = {}) {
        this.data = data;
    }
    via(serverId) {
        return new this.constructor(Object.assign({}, this.data, { serverId }));
    }
    to(target) {
        return new this.constructor(Object.assign({}, this.data, { target }));
    }
    getServerId() {
        let serverId;
        if (this.data.serverId) {
            serverId = this.data.serverId;
            delete this.data.serverId;
        }
        else if (app.rpc.server) {
            serverId = "web-server-1";
        }
        else {
            serverId = app.serverId;
        }
        return serverId;
    }
}
exports.Message = Message;
class WebSocketMessage extends Message {
    constructor() {
        super(...arguments);
        this.name = "app.message.ws";
    }
    get volatile() {
        return new WebSocketMessage(Object.assign({}, this.data, { volatile: true }));
    }
    get local() {
        return new WebSocketMessage(Object.assign({}, this.data, { local: true }));
    }
    of(nsp) {
        return new WebSocketMessage(Object.assign({}, this.data, { nsp }));
    }
    binary(hasBinary) {
        return new WebSocketMessage(Object.assign({}, this.data, { binary: hasBinary }));
    }
    emit(event, ...data) {
        return app.message.publish(this.name, Object.assign({}, this.data, { event,
            data }), [this.getServerId()]);
    }
}
exports.WebSocketMessage = WebSocketMessage;
class SSEMessage extends Message {
    constructor() {
        super(...arguments);
        this.name = "app.message.sse";
    }
    close() {
        return app.message.publish(this.name, Object.assign({}, this.data, { close: true }), [this.getServerId()]);
    }
    send(data) {
        return app.message.publish(this.name, Object.assign({}, this.data, { data }), [this.getServerId()]);
    }
    emit(event, data) {
        return app.message.publish(this.name, Object.assign({}, this.data, { event,
            data }), [this.getServerId()]);
    }
}
exports.SSEMessage = SSEMessage;
//# sourceMappingURL=MessageChannel.js.map