"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const fs = require("fs-extra");
const hash = require("string-hash");
const init_1 = require("../../init");
const module_1 = require("./internal/module");
const moment = require("moment");
class Schedule {
    constructor(name) {
        this.name = name;
    }
    create(options, handler) {
        let { salt, start, end, repeat, module, data } = options;
        let idParam = [app.serverId, module_1.traceModulePath(init_1.ROOT_PATH)];
        let isMethod = false;
        let isCallable = false;
        salt && idParam.push(salt);
        module && idParam.push(module.name);
        handler = handler || options.handler;
        if (handler) {
            if (typeof handler === "string") {
                isMethod = true;
                idParam.push(handler);
            }
            else {
                isCallable = true;
                handler.name && idParam.push(handler.name);
            }
        }
        if (idParam.length < 3) {
            throw new TypeError("not enough options for scheduling, try providing a salt");
        }
        else if (module && typeof handler !== "string") {
            throw new TypeError("handler is missing or not a method of the module");
        }
        if (String(start).length === 13) {
            start = Math.round(start / 1000);
            if (repeat) {
                repeat = Math.ceil(repeat / 1000);
            }
        }
        let taskId = hash(idParam.join("#"));
        let event = String(taskId);
        if (isCallable) {
            app.message.unsubscribe(event);
            app.message.subscribe(event, handler);
        }
        app.services.schedule.instance(event).add({
            taskId,
            serverId: app.serverId,
            start,
            repeat,
            end,
            module: module && module.name,
            handler: isMethod ? handler : void 0,
            data
        });
        return taskId;
    }
    cancel(taskId) {
        if (!taskId)
            return;
        let event = String(taskId);
        app.message.unsubscribe(event);
        app.services.schedule.instance(event).delete(taskId);
    }
}
exports.Schedule = Schedule;
class ScheduleService {
    constructor() {
        this.timer = null;
        this.tasks = new Map();
        this.filename = app.ROOT_PATH + `/cache/schedules-${app.serverId}.json`;
        this.setup().add({
            taskId: -1,
            serverId: app.serverId,
            start: moment().unix(),
            repeat: 60 * 30
        });
    }
    async add(task) {
        this.tasks.set(task.taskId, Object.assign({}, task, { expired: false }));
        return true;
    }
    async delete(taskId) {
        return this.tasks.delete(taskId);
    }
    async stop(transferTasks = false) {
        if (this.state === "stopped") {
            return;
        }
        else {
            this.state = "stopped";
        }
        await this.clear().gc();
        let tasks = [];
        let mod = app.services.schedule;
        let { local } = app;
        for (let [taskId, task] of this.tasks) {
            if (taskId <= 0)
                continue;
            this.tasks.delete(taskId);
            if (transferTasks && mod.instance(taskId) !== mod.instance(local)) {
                mod.instance(taskId).add(task);
            }
            else {
                tasks.push(task);
            }
        }
        if (!transferTasks) {
            await fs.ensureDir(path.dirname(this.filename));
            await fs.writeJSON(this.filename, tasks);
        }
    }
    async resume() {
        this.state === "stopped" && this.setup();
        try {
            let tasks = await fs.readJSON(this.filename);
            for (let task of tasks) {
                Array.isArray(task) && (task = task[1]);
                this.tasks.set(task.taskId, task);
            }
        }
        catch (e) { }
    }
    setup() {
        this.clear().state = "running";
        this.timer = setInterval(() => {
            let now = moment().unix();
            for (let [taskId, task] of this.tasks) {
                let { start, end, expired, serverId, repeat, data } = task;
                if (!expired && now >= start) {
                    if (!end || now < end) {
                        if (taskId === -1 && serverId === app.serverId) {
                            this.gc(now);
                        }
                        else if (!task.handler) {
                            app.message.publish(String(taskId), data, [serverId]);
                        }
                        else {
                            app.message.publish(app.schedule.name, [
                                task.module,
                                task.handler,
                                data
                            ], [serverId]);
                        }
                        if (repeat)
                            task.start = now + task.repeat;
                        else
                            task.expired = true;
                    }
                    else {
                        task.expired = true;
                    }
                }
            }
        }, 1000);
        return this;
    }
    clear() {
        this.timer && clearInterval(this.timer);
        return this;
    }
    async gc(now) {
        now = now || moment().unix();
        for (let [taskId, task] of this.tasks) {
            if (task.expired || (task.end && now >= task.end)) {
                this.tasks.delete(taskId);
            }
        }
    }
}
exports.ScheduleService = ScheduleService;
exports.default = ScheduleService;
//# sourceMappingURL=Schedule.js.map