import * as path from "path";
import * as fs from "fs";
import * as FRON from "fron";
import * as alar from "alar";
import { config as configEnv } from "dotenv";

declare global {
    namespace NodeJS {
        interface Global {
            app: { [name: string]: any };
        }
    }

    namespace app {
        const ROOT_PATH: string;
        const SRC_PATH: string;
        const APP_PATH: string;
        const isDebugMode: boolean;
        const isDevMode: boolean;
        const isTsNode: boolean;
        const isCli: boolean;
        const isWebServer: boolean;

        /**
         * In a web server, the server ID would be `web-server-<n>`, where `<n>`
         * is a synonym of the current worker ID (AKA peer ID, generated by
         * **ipchannel**); in an RPC server, the server ID is the ID passed to
         * `app.rpc.serve()`. NOTE: the server ID will only be available until
         * the server started.
         */
        var serverId: string;

        /**
         * Pass this symbol to `ModuleProxy<any>.instance()` method so that it
         * can always resolve the local instance of the module.
         */
        const local: symbol;
    }
}

var appPath = path.dirname(process.mainModule.filename);
var argv = process.execArgv.join(" ");

appPath === path.resolve(__dirname, "..") && (appPath = __dirname);

/** The version of framework. */
export const version: string = require("../package.json").version;

/** Whether the current process is running in debug/inspect mode. */
export const isDebugMode = argv.includes("inspect") || argv.includes("debug");

/** Whether the current process is running in ts-node. */
export const isTsNode = argv.includes("ts-node");

/** Whether the current process is running in command line. */
export const isCli = appPath == path.resolve(__dirname, "cli");

isCli && (appPath = path.resolve(process.cwd(), "dist"));

/** The root path of the project. */
export const ROOT_PATH = path.normalize(appPath + "/..");

var srcPath: string = appPath;

try {
    let filename = ROOT_PATH + "/tsconfig.json",
        jsonc = fs.readFileSync(filename, "utf8"),
        { compilerOptions } = FRON.parse(jsonc, filename),
        { rootDir, outDir } = compilerOptions || <any>{};

    if (rootDir)
        srcPath = path.normalize(ROOT_PATH + "/" + rootDir);
    if (outDir)
        appPath = path.normalize(ROOT_PATH + "/" + outDir);
} catch (e) {
    srcPath = path.normalize(ROOT_PATH + "/src");
    appPath = path.normalize(ROOT_PATH + "/dist");
}

/** The source code path. */
export const SRC_PATH: string = srcPath;

/** The application path, usually it's the distribution path. */
export const APP_PATH: string = isTsNode ? SRC_PATH : appPath;

/** Whether the program is running in development mode. */
export const isDevMode = isDebugMode || !process.send;

var webEntryFile = path.resolve(APP_PATH, "index" + (isTsNode ? ".ts" : ".js"));

/** Whether the program is running as a web server. */
export const isWebServer = process.mainModule.filename == webEntryFile;

// support .env configuration file
configEnv({ path: ROOT_PATH + "/.env" });

global.app = {
    ROOT_PATH,
    SRC_PATH,
    APP_PATH,
    isDebugMode,
    isDevMode,
    isTsNode,
    isCli,
    isWebServer,
    local: alar.util.local
};